import { assert } from 'chai'
import { cloneDeep } from 'lodash'
import {
  boundingBox,
  pointsEqual,
  shiftPolygon,
  svgPolygon,
  checkPolygon,
  InvalidPolygonError,
  medianPoint
} from '@/helpers'

describe('helpers/polygon', () => {
  describe('pointsEqual', () => {
    it('cannot be replaced by ===', () => {
      assert.notStrictEqual([1, 2], [1, 2])
      assert.notStrictEqual([1, 2], [1, 1])
    })
    it('compares points by value', () => {
      assert(pointsEqual([1, 2], [1, 2]))
      assert(!pointsEqual([1, 2], [1, 1]))
    })
  })

  describe('shiftPolygon', () => {
    it('shifts polygon by an offset', () => {
      const originalPolygon = [[1, 1], [1, 2], [2, 2], [3, 3]]
      const expectedPolygon = [[11, 21], [11, 22], [12, 22], [13, 23]]
      assert.deepStrictEqual(shiftPolygon(originalPolygon, [10, 20]), expectedPolygon)
    })
    it('handles negative offsets', () => {
      const originalPolygon = [[1, 1], [1, 2], [2, 2], [3, 3]]
      const expectedPolygon = [[-9, -19], [-9, -18], [-8, -18], [-7, -17]]
      assert.deepStrictEqual(shiftPolygon(originalPolygon, [-10, -20]), expectedPolygon)
    })
    it('handles no offset', () => {
      const polygon = [[1, 1], [1, 2], [2, 2], [3, 3]]
      assert.deepStrictEqual(shiftPolygon(polygon, [0, 0]), polygon)
    })
  })

  describe('svgPolygon', () => {
    it('turns a polygon into SVG coordinates', () => {
      assert.strictEqual(svgPolygon([]), '')
      assert.strictEqual(svgPolygon([[]]), '')
      assert.strictEqual(svgPolygon([[1]]), '1')
      assert.strictEqual(svgPolygon([[1, 2]]), '1,2')
      assert.strictEqual(svgPolygon([[1, 2], [3, 4], [5, 6]]), '1,2 3,4 5,6')
      assert.strictEqual(
        svgPolygon([[-1.4, 2.2], [-0, +0], [5, 6.7]]),
        '-1.4,2.2 0,0 5,6.7'
      )
    })
  })

  describe('checkPolygon', () => {
    it('checks for argument types', () => {
      assert.throws(
        () => checkPolygon(NaN),
        TypeError,
        'Expected Array, got NaN'
      )
      assert.throws(
        () => checkPolygon([[1, 2, 3]]),
        TypeError,
        'Point 0: expected Array of two finite numbers, got 1,2,3'
      )
      assert.throws(
        () => checkPolygon([[1, 2], [NaN, Infinity]]),
        TypeError,
        'Point 1: expected Array of two finite numbers, got NaN,Infinity'
      )
    })

    it('throws InvalidPolygonError when a polygon is too small', () => {
      assert.throws(
        () => checkPolygon([
          [1, 1],
          [1, 2],
          [2, 2],
          [2, 1]
        ]),
        InvalidPolygonError,
        'This polygon is too small.'
      )
    })

    it('requires three unique points', () => {
      assert.throws(
        () => checkPolygon([[50, 50], [50, 50], [0, 0], [0, 0]]),
        InvalidPolygonError,
        'This polygon does not have at least three unique points.'
      )
    })

    it('requires three unique points, ignoring an equal first and last point', () => {
      assert.throws(
        () => checkPolygon([[50, 50], [50, 50], [0, 0], [0, 0], [50, 50]]),
        InvalidPolygonError,
        'This polygon does not have at least three unique points.'
      )
    })

    it('requires less than 164 unique points', () => {
      assert.doesNotThrow(() => checkPolygon(Array(163).fill().map((value, i) => [i, i])))
      assert.throws(
        () => checkPolygon(Array(200).fill().map((value, i) => [i, i])),
        InvalidPolygonError,
        'This polygon has more than 163 distinct points.'
      )

      assert.throws(
        () => checkPolygon(Array(164).fill().map((value, i) => [i, i])),
        InvalidPolygonError,
        'This polygon has more than 163 distinct points.'
      )
      // 166 points, but the final point is equal to the first point
      assert.doesNotThrow(() => checkPolygon(Array(164).fill().map((value, i) => [i % 163, i % 163])))
    })

    it('requires non-negative coordinates', () => {
      assert.throws(
        () => checkPolygon([[0, 0], [10, 10], [-1, 4], [20, 2], [0, 0]]),
        InvalidPolygonError,
        'A polygon cannot have negative coordinates.'
      )
    })

    it('deduplicates points', () => {
      assert.deepStrictEqual(
        checkPolygon([
          [1, 1],
          [1, 1],
          [50, 50],
          [50, 51],
          [50, 51],
          [50, 50],
          [100, 100]
        ]),
        [
          [1, 1],
          [50, 50],
          [50, 51],
          [50, 50],
          [100, 100]
        ]
      )
    })

    it('rounds floats', () => {
      assert.deepStrictEqual(
        checkPolygon([
          [1.3, 1.7],
          [74.4, 2.9],
          [3.5, 30],
          [1.3, 1.7]
        ]),
        [
          [1, 2],
          [74, 3],
          [4, 30],
          [1, 2]
        ]
      )
    })

    it('does nothing with valid polygons', () => {
      const polygon = [[10, 10], [10, 100], [100, 100], [10, 10]]
      assert.deepStrictEqual(checkPolygon(polygon), polygon)
    })

    it('does not mutate the original polygon', () => {
      const originalPolygon = [
        [1.3, 1.7],
        [74.4, 2.9],
        [74.4, 2.9],
        [3.5, 30],
        [3.5, 30],
        [1.3, 1.7]
      ]
      const expectedPolygon = cloneDeep(originalPolygon)
      checkPolygon(originalPolygon)
      assert.deepStrictEqual(originalPolygon, expectedPolygon)
    })
  })

  describe('boundingBox', () => {
    it("gets a zone's bounding box", () => {
      assert.deepStrictEqual(
        boundingBox({
          image: { width: 100, height: 100 },
          polygon: [
            [30, 20],
            [30, 60],
            [50, 60],
            [50, 20]
          ]
        }),
        { x: 30, y: 20, width: 20, height: 40 }
      )
    })

    it('requires an image', () => {
      assert.throws(
        () => boundingBox({ polygon: [[0, 0], [0, 10], [10, 10], [10, 0], [0, 0]] }),
        Error,
        'An image is required.'
      )
    })

    it('requires an image with valid dimensions', () => {
      assert.throws(
        () => boundingBox({
          image: {},
          polygon: [[0, 0], [0, 10], [10, 10], [10, 0], [0, 0]]
        }),
        Error,
        'An image with valid dimensions is required.'
      )
      assert.throws(
        () => boundingBox({
          image: { width: 42 },
          polygon: [[0, 0], [0, 10], [10, 10], [10, 0], [0, 0]]
        }),
        Error,
        'An image with valid dimensions is required.'
      )
      assert.throws(
        () => boundingBox({
          image: { width: 0, height: 42 },
          polygon: [[0, 0], [0, 10], [10, 10], [10, 0], [0, 0]]
        }),
        Error,
        'An image with valid dimensions is required.'
      )
    })

    it('does not require an image when imageBounds is disabled', () => {
      assert.deepStrictEqual(
        boundingBox({
          polygon: [[0, 0], [0, 10], [10, 10], [10, 0], [0, 0]]
        }, { imageBounds: false }),
        { x: 0, y: 0, width: 10, height: 10 }
      )
    })

    it('prevents negative coordinates', () => {
      assert.deepStrictEqual(
        boundingBox({
          image: { width: 100, height: 100 },
          polygon: [
            [-500, -500],
            [-500, 80],
            [80, 80],
            [80, -500]
          ]
        }),
        { x: 0, y: 0, width: 80, height: 80 }
      )
    })

    it('prevents leaving the image bounds', () => {
      assert.deepStrictEqual(
        boundingBox({
          image: { width: 100, height: 100 },
          polygon: [
            [10, 10],
            [10, 9001],
            [9001, 9001],
            [9001, 10]
          ]
        }),
        { x: 10, y: 10, width: 90, height: 90 }
      )
    })

    it('allows leaving the image bounds without the imageBounds option', () => {
      assert.deepStrictEqual(
        boundingBox({
          image: { width: 100, height: 100 },
          polygon: [
            [10, 10],
            [10, 9001],
            [9001, 9001],
            [9001, 10]
          ]
        }, { imageBounds: false }),
        { x: 10, y: 10, width: 8991, height: 8991 }
      )
    })

    it('can add a margin to a bounding box', () => {
      assert.deepStrictEqual(
        boundingBox({
          image: { width: 100, height: 100 },
          polygon: [
            [30, 20],
            [30, 60],
            [50, 60],
            [50, 20]
          ]
        }, { margin: 15 }),
        { x: 15, y: 5, width: 50, height: 70 }
      )
    })

    it('prevents negative coordinates caused by margins', () => {
      assert.deepStrictEqual(
        boundingBox({
          image: { width: 100, height: 100 },
          polygon: [
            [10, 20],
            [10, 30],
            [20, 30],
            [20, 20]
          ]
        }, { margin: 50 }),
        { x: 0, y: 0, width: 70, height: 80 }
      )
    })

    it('allows margins with negative coordinates without the imageBounds option', () => {
      assert.deepStrictEqual(
        boundingBox({
          image: { width: 100, height: 100 },
          polygon: [
            [10, 20],
            [10, 30],
            [20, 30],
            [20, 20]
          ]
        }, { margin: 50, imageBounds: false }),
        { x: -40, y: -30, width: 110, height: 110 }
      )
    })

    it('prevents leaving the image bounds using margins', () => {
      assert.deepStrictEqual(
        boundingBox({
          image: { width: 100, height: 100 },
          polygon: [
            [80, 70],
            [80, 80],
            [90, 80],
            [90, 70]
          ]
        }, { margin: 50 }),
        { x: 30, y: 20, width: 70, height: 80 }
      )
    })

    it('allows margins leaving the image without the imageBounds option', () => {
      assert.deepStrictEqual(
        boundingBox({
          image: { width: 100, height: 100 },
          polygon: [
            [80, 70],
            [80, 80],
            [90, 80],
            [90, 70]
          ]
        }, { margin: 50, imageBounds: false }),
        { x: 30, y: 20, width: 110, height: 110 }
      )
    })
  })

  describe('medianPoint', () => {
    it('calculates a median point on a polygon line', () => {
      assert.deepStrictEqual(
        medianPoint(
          3,
          [376, 199],
          [
            [315, 199],
            [315, 236],
            [376, 236],
            [376, 199],
            [315, 199]
          ]
        ),
        [346, 199]
      )
    })
  })
})
