import { assert } from 'chai'
import { createPinia, setActivePinia } from 'pinia'
import axios from 'axios'
import { useFolderPickerStore } from '@/stores/folderpicker.ts'
import { assertRejects, FakeAxios } from '../testhelpers.js'

describe('folderpicker', () => {
  let store

  before(() => {
    setActivePinia(createPinia())
    store = useFolderPickerStore()
  })

  beforeEach(() => {
    store.$reset()
  })

  describe('actions', () => {
    let mock

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('nextFolders', () => {
      it('requires a corpus or a folder', async () => {
        const err = await assertRejects(() => store.nextFolders())
        assert.strictEqual(err.message, 'Either a corpus ID or a folder ID are required')
      })

      it('loads multiple pages', async () => {
        const folder1 = { id: 'folder1', name: 'Folder 1' }
        const folder2 = { id: 'folder2', name: 'Folder 2' }
        const folder3 = { id: 'folder3', name: 'Folder 3' }

        store.folders = { folder1 }

        const pagination1 = {
          count: 3,
          previous: null,
          next: '/elements/folder1/children/?page=2',
          number: 1
        }
        const pagination2 = {
          count: 3,
          previous: '/elements/folder1/children/?page=1',
          next: '/elements/folder1/children/?page=3',
          number: 2
        }

        mock.onGet('/elements/folder1/children/', { params: { folder: true, page: 1 } }).reply(200, {
          ...pagination1,
          results: [folder2]
        })
        mock.onGet('/elements/folder1/children/', { params: { folder: true, page: 2 } }).reply(200, {
          ...pagination2,
          results: [folder3]
        })

        await store.nextFolders('corpusid', 'folder1', 2)

        assert.deepStrictEqual(store.folders, {
          folder1: {
            ...folder1,
            subfolders: ['folder2', 'folder3']
          },
          folder2: {
            ...folder2,
            subfolders: []
          },
          folder3: {
            ...folder3,
            subfolders: []
          }
        })

        assert.deepStrictEqual(store.folderPagination, {
          folder1: pagination2
        })
      })

      it('accepts a folder with no corpus', async () => {
        const folder1 = { id: 'folder1', name: 'Folder 1' }
        const folder2 = { id: 'folder2', name: 'Folder 2' }
        const folder3 = { id: 'folder3', name: 'Folder 3' }

        store.folders = { folder1 }

        const pagination1 = {
          count: 3,
          previous: null,
          next: '/elements/folder1/children/?page=2',
          number: 1
        }
        const pagination2 = {
          count: 3,
          previous: '/elements/folder1/children/?page=1',
          next: '/elements/folder1/children/?page=3',
          number: 2
        }

        mock.onGet('/elements/folder1/children/', { params: { folder: true, page: 1 } }).reply(200, {
          ...pagination1,
          results: [folder2]
        })
        mock.onGet('/elements/folder1/children/', { params: { folder: true, page: 2 } }).reply(200, {
          ...pagination2,
          results: [folder3]
        })

        await store.nextFolders(null, 'folder1', 2)

        assert.deepStrictEqual(store.folders, {
          folder1: {
            ...folder1,
            subfolders: ['folder2', 'folder3']
          },
          folder2: {
            ...folder2,
            subfolders: []
          },
          folder3: {
            ...folder3,
            subfolders: []
          }
        })

        assert.deepStrictEqual(store.folderPagination, {
          folder1: pagination2
        })
      })

      it('uses listElements for a corpus', async () => {
        const folder1 = { id: 'folder1', name: 'Folder 1' }
        const folder2 = { id: 'folder2', name: 'Folder 2' }

        const pagination1 = {
          count: 3,
          previous: null,
          next: '/corpus/corpusid/elements/?page=2',
          number: 1
        }
        const pagination2 = {
          count: 3,
          previous: '/corpus/corpusid/elements/?page=1',
          next: '/corpus/corpusid/elements/?page=3',
          number: 2
        }

        mock.onGet('/corpus/corpusid/elements/', {
          params: {
            folder: true,
            top_level: true,
            page: 1
          }
        }).reply(200, {
          ...pagination1,
          results: [folder1]
        })
        mock.onGet('/corpus/corpusid/elements/', {
          params: {
            folder: true,
            top_level: true,
            page: 2
          }
        }).reply(200, {
          ...pagination2,
          results: [folder2]
        })

        await store.nextFolders('corpusid', null, 2)

        assert.deepStrictEqual(store.folders, {
          folder1: {
            ...folder1,
            subfolders: []
          },
          folder2: {
            ...folder2,
            subfolders: []
          }
        })
        assert.deepStrictEqual(store.corpus, {
          corpusid: ['folder1', 'folder2']
        })

        assert.deepStrictEqual(store.corpusPagination, {
          corpusid: pagination2
        })
        assert.deepStrictEqual(store.folderPagination, {})
      })

      it('does nothing when there are no further pages', async () => {
        const initialFolderPagination = {
          folder1: {
            count: 0,
            previous: null,
            next: null,
            number: 1
          }
        }
        store.folderPagination = initialFolderPagination

        await store.nextFolders(null, 'folder1', Infinity)

        assert.strictEqual(mock.history.get.length, 0)
        assert.deepStrictEqual(store.folderPagination, initialFolderPagination)
      })

      it('handles errors', async () => {
        const folder1 = { id: 'folder1', name: 'Folder 1' }
        const folder2 = { id: 'folder2', name: 'Folder 2' }

        store.folders = { folder1 }

        const pagination1 = {
          count: 3,
          previous: null,
          next: '/elements/folder1/children/?page=2',
          number: 1
        }

        mock.onGet('/elements/folder1/children/', { params: { folder: true, page: 1 } }).reply(200, {
          ...pagination1,
          results: [folder2]
        })
        mock.onGet('/elements/folder1/children/', { params: { folder: true, page: 2 } }).reply(500)

        await assertRejects(() => store.nextFolders('corpusid', 'folder1', 2))

        assert.deepStrictEqual(store.folders, {
          folder1: {
            ...folder1,
            subfolders: ['folder2']
          },
          folder2: {
            ...folder2,
            subfolders: []
          }
        })

        assert.deepStrictEqual(store.folderPagination, {
          folder1: pagination1
        })
      })
    })
  })

  describe('getters', () => {
    describe('pagination', () => {
      it('retrieves pagination for a corpus', () => {
        store.corpusPagination.corpusid = {
          number: 4,
          next: 'url'
        }
        assert.deepStrictEqual(store.pagination('corpusid', null), {
          number: 4,
          next: 'url'
        })
        assert.strictEqual(store.pagination('corpus2', null), null)
      })

      it('retrieves pagination for a folder', () => {
        store.folderPagination.folderid = {
          number: 4,
          next: 'url'
        }
        assert.deepStrictEqual(store.pagination(null, 'folderid'), {
          number: 4,
          next: 'url'
        })
        assert.deepStrictEqual(store.pagination('corpusid', 'folderid'), {
          number: 4,
          next: 'url'
        })
        assert.strictEqual(store.pagination('corpusid', 'folder2'), null)
      })
    })
  })
})
