import { assert } from 'chai'
import axios from 'axios'
import { createPinia, setActivePinia } from 'pinia'
import { isFunction } from 'lodash'
import { jobsSample } from '../samples'
import { FakeAxios } from '../testhelpers'
import sinon from 'sinon'
import { JOBS_POLLING_DELAY } from '@/config'
import { useJobsStore, useNotificationStore } from '@/stores'

describe('jobs', () => {
  let mock, store, notificationStore

  before('Setting up mocks', () => {
    mock = new FakeAxios(axios)
    setActivePinia(createPinia())
    store = useJobsStore()
    notificationStore = useNotificationStore()
  })

  beforeEach(() => {
    // Remove any handlers, but leave mocking in place
    mock.reset()
    store.$reset()
    notificationStore.$reset()
  })

  after('Removing Axios mock', () => {
    // Remove mocking entirely
    mock.restore()
  })

  describe('list', () => {
    it('lists jobs', async () => {
      mock.onGet('/jobs/').reply(() => {
        // While the API request is running, the store is loading
        assert.isTrue(store.loading)
        return [200, jobsSample]
      })

      await store.list()

      assert.deepStrictEqual(store.jobs, { jobid: jobsSample[0] })
      assert.isFalse(store.loading)
    })

    it('handles errors', async () => {
      mock.onGet('/jobs/').reply(418, { detail: 'no can do, am teapot' })

      await store.list()

      assert.deepStrictEqual(store.jobs, {})
      assert.deepStrictEqual(notificationStore.notifications, [
        {
          id: 0,
          type: 'error',
          text: 'no can do, am teapot'
        }
      ])
    })
  })

  describe('delete', () => {
    it('deletes a job', async () => {
      store.jobs = {
        job1: jobsSample[0],
        job2: { id: 'job2' }
      }
      mock.onDelete('/jobs/job1/').reply(204)

      await store.delete('job1')

      assert.deepStrictEqual(store.jobs, {
        job2: { id: 'job2' }
      })
    })

    it('handles errors', async () => {
      store.jobs = { jobid: jobsSample[0] }
      mock.onDelete('/jobs/jobid/').reply(418, { detail: 'no can do, am teapot' })

      await store.delete('jobid')

      assert.deepStrictEqual(store.jobs, {
        jobid: jobsSample[0]
      })
      assert.deepStrictEqual(notificationStore.notifications, [
        {
          id: 0,
          type: 'error',
          text: 'no can do, am teapot'
        }
      ])
    })
  })

  describe('startPolling', () => {
    it('starts jobs list polling', async () => {
      const oldSetTimeout = setTimeout
      global.setTimeout = sinon.mock('setTimeout').twice()
      setTimeout.callsFake(() => 'timeoutid')

      mock.onGet('/jobs/').reply(200, jobsSample)

      await store.startPolling()
      const [func, delay] = setTimeout.lastCall.args
      assert.ok(isFunction(func))
      assert.strictEqual(delay, JOBS_POLLING_DELAY)

      assert.deepStrictEqual(store.jobs, { jobid: jobsSample[0] })
      assert.strictEqual(store.timeoutId, 'timeoutid')
      assert.isTrue(store.polling)
      assert.isFalse(store.loading)

      // Calling the timeouted function a second time should list jobs again
      store.$reset()
      store.polling = true
      await func()

      assert.deepStrictEqual(store.jobs, { jobid: jobsSample[0] })
      assert.strictEqual(store.timeoutId, 'timeoutid')
      assert.isTrue(store.polling)
      assert.isFalse(store.loading)

      global.setTimeout.verify()
      global.setTimeout = oldSetTimeout
    })
  })

  describe('stopPolling', () => {
    let spy

    before('set up clearTimeout spy', () => {
      spy = sinon.spy(global, 'clearTimeout')
    })

    after('tear down clearTimeout spy', () => {
      spy.restore()
    })

    it('stops jobs list polling', () => {
      store.jobs = { jobid: jobsSample[0] }
      store.timeoutId = 'timeoutid'
      store.polling = true
      store.loading = false

      store.stopPolling()
      assert.ok(spy.calledOnceWith('timeoutid'))

      assert.deepStrictEqual(store.jobs, { jobid: jobsSample[0] })
      assert.isNull(store.timeoutId)
      assert.isFalse(store.polling)
      assert.isFalse(store.loading)
    })
  })
})
