import { assert } from 'chai'
import axios from 'axios'
import { pick } from 'lodash'
import { createPinia, setActivePinia } from 'pinia'
import { usePonosStore } from '@/stores'
import { agentSample, taskSample, farmsSample } from '../samples'
import { FakeAxios } from '../testhelpers'

describe('ponos', () => {
  describe('actions', () => {
    let mock, store

    before('Setting up mocks', () => {
      mock = new FakeAxios(axios)
      setActivePinia(createPinia())
      store = usePonosStore()
    })

    afterEach(() => {
      // Remove any handlers, but leave mocking in place
      mock.reset()
      store.$reset()
    })

    after('Removing Axios mock', () => {
      // Remove mocking entirely
      mock.restore()
    })

    describe('retrieveAgent', () => {
      it('retrieve information about a Ponos agent', async () => {
        store.agents = { [agentSample.id]: { ...agentSample } }

        const payload = { ...agentSample, running_tasks: [taskSample] }
        mock.onGet('/agent/agentid/').reply(200, payload)

        await store.retrieveAgent('agentid')
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'get',
            url: '/agent/agentid/',
            data: undefined
          }
        ])
        assert.deepStrictEqual(store.agents, {
          [agentSample.id]: {
            ...agentSample,
            running_tasks: [taskSample]
          }
        })
      })
    })

    describe('listAgents', () => {
      it('lists all Ponos agents', async () => {
        store.agents = {}

        mock.onGet('/agents/', { page: 1 }).reply(200, {
          count: 1,
          number: 1,
          next: null,
          results: [agentSample]
        })

        await store.listAgents()
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'get',
            url: '/agents/',
            data: undefined
          }
        ])
        assert.deepStrictEqual(store.agents, { [agentSample.id]: agentSample })
      })
    })

    describe('listFarms', () => {
      it('lists all Ponos farms', async () => {
        store.farms = null

        mock.onGet('/farms/', { page: 1 }).reply(200, { ...farmsSample, next: null })

        await store.listFarms()
        assert.deepStrictEqual(mock.history.all.map(req => pick(req, ['method', 'url', 'data'])), [
          {
            method: 'get',
            url: '/farms/',
            data: undefined
          }
        ])
        assert.deepStrictEqual(store.farms, {
          farm1: farmsSample.results[0],
          farm2: farmsSample.results[1]
        })
      })
    })
  })
})
